static  const char rcsid[] = "$Id: bent.c,v 1.9 2004/04/30 09:34:06 dengxf Exp $";

/*
 * Copyright (c) 2003-2004 Deng XueFeng <dsnofe@hotmail.com>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *    This product includes software developed by Deng XueFeng.
 * 4. Neither the name of the author nor the names of any co-contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY DENG XUEFENG AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */


#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <pthread.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <string.h>

#define BUFLEN  255
#define NUMT  4096

char    rhost[32];
int     rport;
char    ruser[32];
char    rpass[32];
char    ocmd[16];
int     btimes;


void
thread(void *arg)
{
    int     i;
    int     mtm = 50000;
    struct sockaddr_in servaddr;
    int     sockfd, n;
    int    t_id = *(int *) arg;
    char    resmsg[BUFLEN + 1];
    char    quitmsg[] = "quit \r\n";

    printf("THREAD:[%d] START.\n", t_id);
    sleep(1);
    sockfd = socket(AF_INET, SOCK_STREAM, 0);
    if (sockfd < 0) {
        printf("THREAD:[%d]:socket creating error.\n", t_id);
        pthread_exit(NULL);
    }
    memset(resmsg, 0, (BUFLEN + 1));
    memset(&servaddr, 0, sizeof(struct sockaddr_in));
    servaddr.sin_family = AF_INET;
    servaddr.sin_port = htons(rport);
    if (inet_pton(AF_INET, rhost, &servaddr.sin_addr) <= 0) {
        printf("THREAD:[%d]:Wrong IP Address.\n", t_id);
        pthread_exit(NULL);
    }
    if (connect(sockfd, (struct sockaddr *) (&servaddr), sizeof(struct sockaddr_in)) < 0) {
        printf("THREAD:[%d]:Connect Failed.\n", t_id);
        pthread_exit(NULL);
    }
    printf("THREAD[%d] Connect to server.\n", t_id);
    n = read(sockfd, resmsg, (BUFLEN + 1));
    if (n <= 0) {
        printf("THREAD:[%d]:socket read err.\n", t_id);
        pthread_exit(NULL);
    }
    resmsg[n] = '\0';
    if ((strstr(resmsg, "OK")) == NULL) {
        pthread_exit(NULL);
    }
    memset(resmsg, 0, strlen(resmsg));

    sleep(1);
    write(sockfd, ruser, strlen(ruser));
    n = read(sockfd, resmsg, (BUFLEN + 1));
    if (n <= 0) {
        printf("THREAD:[%d][%d]:socket read err.\n", t_id, sockfd);
        pthread_exit(NULL);
    }
    resmsg[n] = '\0';
    if ((strstr(resmsg, "OK")) == NULL) {
        pthread_exit(NULL);
    }
    sleep(1);
    write(sockfd, rpass, strlen(rpass));
    n = read(sockfd, resmsg, (BUFLEN + 1));
    if (n <= 0) {
        printf("THREAD:[%d][%d]:socket read err.\n", t_id, sockfd);
        pthread_exit(NULL);
    }
    resmsg[n] = '\0';
    if ((strstr(resmsg, "OK")) == NULL) {
        pthread_exit(NULL);
    }
    sleep(1);
    i = 0;
    while (i < btimes) {
        memset(resmsg, 0, strlen(resmsg));
        write(sockfd, ocmd, strlen(ocmd));
        n = read(sockfd, resmsg, (BUFLEN + 1));
        if (n <= 0) {
            printf("THREAD:loop:read:[%d]:.\n", t_id);
            pthread_exit(NULL);
        }
        resmsg[n] = '\0';
        if ((strstr(resmsg, "OK")) == NULL) {
            printf("THREAD:Optional comand:err:[%d][%d]:.\n", t_id, sockfd);
            pthread_exit(NULL);
        }
        printf("THREAD:Optional command:OK:[%d][%d][%d]:.\n", t_id, sockfd, i);
        i++;
        usleep(mtm);
    }
    write(sockfd, quitmsg, strlen(quitmsg));
}

void
usage(void)
{
    printf("Usage:\nbent -h[Ip] -p[port] -u[user] -s[password] -n[looptimes] -c[optional command] -t[max threads]\n");
}

int
main(int argc, char *argv[])
{
    pthread_t tid[NUMT];
    int     k;
    void   *thread_result;
    int     res;
    int     total;
    int     ntm = 10000;
    char    ch;

    memset(rhost, 0, sizeof(rhost));
    rport = 0;
    memset(ruser, 0, sizeof(ruser));
    memset(rpass, 0, sizeof(rpass));
    memset(ocmd, 0, sizeof(ocmd));
    btimes = 0;
    total = 0;

    while ((ch = getopt(argc, argv, "h:p:u:s:n:c:t:")) != -1) {
        switch (ch) {
            case 'h':
                strlcpy(rhost, optarg, sizeof(rhost));
                break;
            case 'p':
                rport = atoi(optarg);
                break;
            case 'u':
                snprintf(ruser, sizeof(ruser), "user %s\r\n", optarg);
                break;
            case 's':
                snprintf(rpass, sizeof(rpass), "pass %s\r\n", optarg);
                break;
            case 'n':
                btimes = atoi(optarg);
                break;
            case 'c':
                snprintf(ocmd, sizeof(ocmd), "%s\r\n", optarg);
                break;
            case 't':
                total = atoi(optarg);
                break;
            default :
                break;
        }
    }
    if ((strlen(rhost)) == 0 || (strlen(ruser)) == 0 || (strlen(rpass)) == 0 || (strlen(ocmd)) == 0) {
        usage();
        exit(1);
    }
    if (rport == 0 || btimes == 0 || total == 0) {
        usage();
        exit(1);
    }
    printf("Start threads[%d]\n", total);
    sleep(2);

    for (k = 0; k < total; k++) {
        res = pthread_create(&(tid[k]), NULL, (void *)thread, (void *)&k);
        if (res != 0) {
            perror("Create pthread error!\n");
            exit(0);;
        }
        usleep(ntm);
    }
    for (k = total - 1; k >= 0; k--) {
        res = pthread_join(tid[k], &thread_result);
        if (res == 0) {
            printf("Picked a thread[%d]\n", k);
        } else {
            perror("pthread_join failed.\n");
        }
    }
    printf("All done.\n");
    exit(0);
}
