static  const char rcsid[] = "$Id: cfg.c,v 1.3 2005/05/04 01:52:03 dengxf Exp $";

/*
 * Copyright (c) 2003-2004 Deng XueFeng <dsnofe@hotmail.com>
 * All rights reserved.
 */

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <unistd.h>
#include <string.h>
#include <sys/mman.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <errno.h>

#include "mylog.h"
#include "cfg.h"

static  CFG_NODE *cfg_node = NULL;
static  int cfg_node_size = 0;
static  int cfg_node_used = 0;

extern int errno;

static int add_conf(const char *start, const char *const end)
{
    char *p, vb;
    CFG_NODE *p_cfg_node;

    if ((start == NULL) || (end == NULL))
        return (-1);
    for (; (start < end) && ((*start == ' ') || (*start == '\t')); start++);
    if (*start == '#')
        return (-1);
    if (start + 3 >= end)
        return (-1);
    if (cfg_node_size == cfg_node_used) {
        cfg_node_size += 32;
        p_cfg_node = (CFG_NODE *) realloc(cfg_node, cfg_node_size * sizeof(CFG_NODE));
        if (p_cfg_node == NULL)
            return (-2);
        cfg_node = p_cfg_node;
        memset(cfg_node + cfg_node_size - 32, 0, 32 * sizeof(CFG_NODE));
    }
    p_cfg_node = cfg_node + cfg_node_used;
    p = p_cfg_node->key;
    for (; (start < end) && (*start != ' ') && (*start != '\t'); *p++ = *start++);
    if (start == end) {
        memset(p_cfg_node, 0, sizeof(CFG_NODE));
        return (-1);
    }
    for (; (start < end) && ((*start == ' ') || (*start == '\t')); start++);
    if ((start < end) && (*start == '='))
        for (++start; (start < end) && ((*start == ' ') || (*start == '\t')); start++);
    p = p_cfg_node->value;
    if ((*start == '\"') || (*start == '\'')) {
        vb = *start++;
        if (start == end)
            return (-1);
        for (; (start < end) && (*start != vb); *p++ = *start++);
    } else
        for (; (start < end) && (*start != ' ') && (*start != '\t'); *p++ = *start++);
    p_cfg_node->key[CFG_KEY_LEN - 1] = '\0';
    p_cfg_node->value[CFG_VALUE_LEN - 1] = '\0';
    cfg_node_used++;
    return (0);
}

int init_conf(const char *cfg_file)
{
    struct stat stat_buf;
    char *save;
    char *start;
    char *end;
    int fd_cfg_file;

    fd_cfg_file = open(cfg_file, O_RDONLY);
    if (fd_cfg_file == -1) {
        INFO("[-ERR]: open(2) %s: %s\n", cfg_file, strerror(errno));
        return (-1);
    }
    if ((stat(cfg_file, &stat_buf)) == -1) {
        INFO("[-ERR]: stat(2) %s: %s\n", cfg_file, strerror(errno));
        close(fd_cfg_file);
        return (-1);
    }
    save = mmap(0, stat_buf.st_size, PROT_READ|PROT_WRITE, MAP_PRIVATE, fd_cfg_file, 0L);
    close(fd_cfg_file);
    if (save == MAP_FAILED) {
        INFO("[-ERR]: mmap(2) %s: %s\n", cfg_file, strerror(errno));
        return (-1);
    }
    start = save;
    end = save;
    while (end < stat_buf.st_size + save) {
        if ((*end == '\r') || (*end == '\n')) {
            if (add_conf(start, end) == -2) {
                close_conf();
            }
            end++;
            if ((end < stat_buf.st_size + save) && (*end == '\n'))
                end++;   
            start = end;
        } else
            end++;
    }
    if (start < end) {
        if (add_conf(start, end) == -2)
            close_conf();
    }
    munmap(save, stat_buf.st_size);
    return (0);
}

char *get_conf(const char *fmt, ...)
{
    CFG_NODE *p_cfg_node;
    char name[CFG_KEY_LEN];
    va_list args;

    if (cfg_node_used == 0)
        return (NULL);
    va_start(args, fmt);
    if (vsnprintf(name, sizeof(name), fmt, args) >= (int)sizeof(name))
        return (NULL);
    va_end(args);
    name[CFG_KEY_LEN-1] = '\0';
    p_cfg_node = cfg_node;
    while (p_cfg_node < cfg_node + cfg_node_used) {
        if ((strcmp(name, p_cfg_node->key) == 0) && (p_cfg_node->value[0] != 0x00))
            return (p_cfg_node->value);
        p_cfg_node++;
    }
    return (NULL);
}

void close_conf(void)
{
    if (cfg_node != NULL) {
        free(cfg_node);
        cfg_node = NULL;
        cfg_node_size = 0;
        cfg_node_used = 0;
    }
}
