static  const char rcsid[] = "$Id: listen.c,v 1.8 2005/05/04 01:52:03 dengxf Exp $";

/*
 * Copyright (c) 2003-2004 Deng XueFeng <dsnofe@hotmail.com>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *    This product includes software developed by Deng XueFeng.
 * 4. Neither the name of the author nor the names of any co-contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY DENG XUEFENG AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */

#include <sys/types.h>
#include <sys/socket.h>
#include <sys/uio.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <errno.h>
#include <unistd.h>
#include <string.h>

#include "listen.h"

extern int errno;

int newlisten(int *err, const char *const s_ip, const int port, const int linger, const int backlog)
{
    struct sockaddr_in address;
    struct in_addr ip;
    struct linger lin;
    socklen_t tmp;
    int fd;

    if (port == 0)
        return (ERROR_PORT);
    memset(&address, 0, sizeof(address));
    if (*s_ip == '\0')
        address.sin_addr.s_addr = htonl(INADDR_ANY);
    else {
        if (inet_aton(s_ip, &ip) != 1)
            return (ERROR_IP);
        address.sin_addr.s_addr = ip.s_addr;
    }
    address.sin_port = htons(port);
    address.sin_family = AF_INET;
    fd = socket(AF_INET, SOCK_STREAM, 0);
    if (fd < 0) {
        *err = errno;
        return (ERROR_SOCKET);
    }
    tmp = 1;
    if (setsockopt(fd, SOL_SOCKET, SO_REUSEADDR, &tmp, sizeof(tmp)) < 0) {
        *err = errno;
        close(fd);
        return (ERROR_REUSE);
    }
    lin.l_onoff = 1;
    lin.l_linger = linger;
    if (setsockopt(fd, SOL_SOCKET, SO_LINGER, &lin, sizeof(struct linger)) < 0) {
        *err = errno;
        close(fd);
        return (ERROR_LINGER);
    }
    if (bind(fd, (struct sockaddr *) & address, sizeof(struct sockaddr_in)) < 0) {
        *err = errno;
        close(fd);
        return (ERROR_BIND);
    }
    if (listen(fd, backlog) < 0) {
        *err = errno;
        close(fd);
        return (ERROR_LISTEN);
    }
    return (fd);
}
