static  const char rcsid[] = "$Id: main.c,v 1.20 2005/05/04 03:56:40 dengxf Exp $";

/*
 * Copyright (c) 2003-2004 Deng XueFeng <dsnofe@hotmail.com>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *    This product includes software developed by Deng XueFeng.
 * 4. Neither the name of the author nor the names of any co-contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY DENG XUEFENG AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <signal.h>
#include <syslog.h>
#include <stdarg.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/wait.h>
#include <sys/resource.h>

#include "plcat.h"
#include "mylog.h"
#include "listen.h"
#include "pop3d.h"
#include "mydaemon.h"
#include "cfg.h"
#include "sig.h"

static int quit_flag = 0;

static char config_file[] = "/usr/local/etc/mps/pop3d.conf";
static char pid_file[] = "/var/run/mps.pid";
static int *pop3dpid = NULL;

int pop3d_port = 110;
int max_errs = 5;
int multi_process = 5;
int max_backlog = 32;
int max_client = 1000;
int max_timeout = 10 * 1000;
char serv_ip[16] = "";

int debug_flag = 0;
int pop3d_fd = 0;
int cur_process = 0;

extern int errno;

static void close_child(void)
{
    kill(0, SIGINT);
}

static void sig_stop_handler(int signo)
{
    quit_flag = 1;
    INFO("[INFO]: Received SIG%s and EXIT\n", sigs[signo-1]);
    close_child();
}

static void sig_exit_handler(int signo)
{
    INFO("[INFO]: Received SIG%s and EXIT\n", sigs[signo-1]);
    exit(EXIT_FAILURE);
}

static void sig_handler(int signo)
{
    INFO("[INFO]: Received SIG%s\n", sigs[signo-1]);
}

static int pop3d_child(int id)
{
    int pid;

    pid = fork();
    if (pid < 0) {
        INFO("[INFO]: fork(2) Failed first time: %s", strerror(errno));
        return (0);
    } else if (pid == 0) {
        signal(SIGINT, sig_exit_handler);
        sleep(1);
        pop3d(id);
        return (1);
    } else {
        pop3dpid[id] = pid;
        cur_process++;
        return (1);
    }
}

static int get_id(int pid)
{
    int i;

    i = 0;
    while (i < multi_process) {
        if (pop3dpid[i] == pid)
            return (i);
        i++;
    }
    INFO("[EXIT]: get_pid: failed get pid: %d\n", pid);
    close_child();
    return (multi_process);
}

void sig_chld_handler(int signo)
{
    int id;
    pid_t pid;

    INFO("[INFO]: Received SIG%s\n", sigs[signo-1]);
    while((pid = waitpid(-1, NULL, WNOHANG)) > 0) {
        cur_process--;
        if ((quit_flag == 0) && (debug_flag == 0)) {
            id = get_id(pid);
            pop3d_child(id);
        }
    }
}

static int parse_conf(const char *cfg_file)
{
    char *value;
    
    if (init_conf(cfg_file) != 0)
        return (-1);
    if ((value = get_conf("Client_Max_Errors")) != NULL)
        max_errs = atoi(value);
    if ((value = get_conf("Client_Time_Out")) != NULL)
        max_timeout = 1000 * atoi(value);
    if ((value = get_conf("Per_Process_Max_Client")) != NULL)
        max_client = atoi(value);
    if ((value = get_conf("Multi_Process")) != NULL)
        multi_process = atoi(value);
    if ((value = get_conf("Max_Backlog")) != NULL)
        max_backlog = atoi(value);
    if ((value = get_conf("Server_Port")) != NULL)
        pop3d_port = atoi(value);
    if ((value = get_conf("Server_Address")) != NULL)
        strncpy(serv_ip, value, sizeof(serv_ip));
    if ((value = get_conf("Pid_File")) != NULL)
        strncpy(pid_file, value, sizeof(pid_file));
    close_conf();
    return (0);
}

static void delpid(void)
{
    unlink(pid_file);
}

static int savepid(void)
{
    int fd;
    int pid;
    char str[12];

    fd = open(pid_file, O_WRONLY|O_CREAT|O_TRUNC, 0644);
    if (fd < 0)
        return (errno);
    pid = getpid();
    sprintf(str, "%d\n", pid);
    if (write(fd, str, strlen(str)) != (int)strlen(str)) {
        pid = errno;
        close(fd);
        return (pid);
    }
    close(fd);
    return (0);
}

static int checkpid(void)
{
    int fd;
    int res;
    int pid;
    char str[12];

    fd = open(pid_file, O_RDONLY);
    if (fd < 0)
        return (-1);
    if ((res = read(fd, str, sizeof(str))) <= 0) {
        close(fd);
        return (-1);
    }
    close(fd);
    str[res - 1] = 0x00;
    pid = atoi(str);
    res = kill(pid, 0);
    if (res != 0) {
        INFO("[-ERR]: kill(2) Failed: %s\n", strerror(errno));
        return (-1);
    }
    return (pid);
}

static void myexit(void)
{
    int res;

    res = checkpid();
    if (res > 0) {
        kill(res, SIGUSR1);
        printf("[INFO] POP3D IS STOPPED\n");
    } else
        printf("[INFO] POP3D IS NOT RUNNING\n");
}

static void usage(void)
{
    printf("Usage: mps [-c configfile] [-d] [-q]\n");
    printf("       -c   Specified configfile, default is /usr/local/etc/mps/pop3d.conf\n");
    printf("       -d   Debug Mod\n");
    printf("       -q   Quit\n");
    exit(EXIT_FAILURE);
}

int main(int argc, char **argv)
{
    char opt;
    char cfg_file[256];
    int err;
    int i;

    (void) memset(cfg_file, 0, sizeof(config_file));
    while ((opt = getopt(argc, argv, "c:dq")) != -1) {
        switch(opt) {
            case 'c':
                strlcpy(cfg_file, optarg, sizeof(cfg_file));
                cfg_file[255] = '\0';
                break;
            case 'd':
                debug_flag = 1;
                break;
            case 'q':
                quit_flag = 1;
                break;
            default :
                usage();
        }
    }
    if (cfg_file[0] == '\0')
        strlcpy(cfg_file, config_file, sizeof(cfg_file));
    if (parse_conf(cfg_file) != 0)
        usage();
    if (quit_flag == 1) {
        myexit();
        exit(EXIT_SUCCESS);
    }
    if (checkpid() > 0) {
        printf("[INFO] POP3D IS RUNNING NOW\n");
        exit(EXIT_SUCCESS);
    }
    if (atexit(delpid) != 0) {
        INFO("[INFO]: Register delpid Failed: %s\n", strerror(errno));
        exit(EXIT_SUCCESS);
    }
    delpid();
    if (debug_flag == 1)
        multi_process = 1;
    else
        mydaemon();
    if ((i = savepid()) != 0) {
        INFO("[INFO]: savepid() Failed: %s\n", strerror(i));
        exit(EXIT_FAILURE);
    } else
        INFO("[INFO]: savepid() Successful\n");

    pop3dpid = malloc((multi_process+1) * sizeof(int));
    if (pop3dpid == NULL) {
        INFO("[-ERR]: malloc(3): %s\n", strerror(errno));
        exit(EXIT_FAILURE);
    }
    memset(pop3dpid, 0, (multi_process+1) * sizeof(int));

    pop3d_fd = newlisten(&err, serv_ip, pop3d_port, 10, max_backlog);
    switch (pop3d_fd) {
        case ERROR_PORT:
            INFO("-ERR INVALID PORT: [%d].\n", pop3d_port);
            exit(EXIT_FAILURE);
        case ERROR_IP:
            INFO("-ERR INVALID IP ADDRESS: [%s].\n", serv_ip);
            exit(EXIT_FAILURE);
        case ERROR_SOCKET:
            INFO("-ERR CREATE SOCKET FAILED: [%s].\n", strerror(err));
            exit(EXIT_FAILURE);
        case ERROR_REUSE:
            INFO("-ERR SETSOCKOPT REUSEADDR FAILED: [%s].\n", strerror(err));
            exit(EXIT_FAILURE);
        case ERROR_LINGER:
            INFO("-ERR SETSOCKOPT LINGER FAILED: [%s].\n", strerror(err));
            exit(EXIT_FAILURE);
        case ERROR_BIND:
            INFO("-ERR BIND SOCKET FAILED: [%s].\n", strerror(err));
            exit(EXIT_FAILURE);
        case ERROR_LISTEN:
            INFO("-ERR LISTEN SOCKET FAILED: [%s].\n", strerror(err));
            exit(EXIT_FAILURE);
        default:
            break;
    }
    signal(SIGILL, sig_exit_handler);
    signal(SIGABRT, sig_exit_handler);
    signal(SIGFPE, sig_exit_handler);
    signal(SIGBUS, sig_exit_handler);
    signal(SIGPIPE, sig_handler);
    signal(SIGTERM, sig_stop_handler);
    signal(SIGQUIT, sig_stop_handler);
    signal(SIGINT, sig_handler);
    signal(SIGUSR1, sig_stop_handler);
    signal(SIGUSR2, sig_stop_handler);
    signal(SIGCHLD, sig_chld_handler);
    if (debug_flag == 1) {
        signal(SIGTERM, sig_exit_handler);
        signal(SIGQUIT, sig_exit_handler);
        signal(SIGINT, sig_exit_handler);
        signal(SIGUSR1, sig_exit_handler);
        signal(SIGUSR2, sig_exit_handler);
        pop3d(0);
        return (0);
    }

    i = 0;
    while (i < multi_process) {
        if (pop3d_child(i) == 0) {
            close_child();
            goto WAIT_TO_EXIT;
        }
        i++;
    }
WAIT_TO_EXIT:
    while(1) {
        sleep(30);
        if (cur_process == 0)
            break;
    }
    close(pop3d_fd);
    INFO("EXIT SUCCESSFUL\n");
    return (0);
}
