static  const char rcsid[] = "$Id: plcat.c,v 1.7 2005/05/04 01:52:04 dengxf Exp $";

/*
 * Copyright (c) 2003-2004 Deng XueFeng <dsnofe@hotmail.com>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *    This product includes software developed by Deng XueFeng.
 * 4. Neither the name of the author nor the names of any co-contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY DENG XUEFENG AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */

#include <stdarg.h>
#include <string.h>

int itoa(char *buf, int num)
{
    int res;
    int div;
    int flag;
    int len;

    len = flag = 0;
    if (num < 0) {
        *buf++ = '-';
        num = -num;
        len++;
    }
    if (num == 0) {
        *buf++ = '0';
        len++;
    } else {
        div = (num < 100000) ? 100000 : 1000000000;
        while (div > 0) {
            res = num / div;
            if ((flag == 1) || (res > 0)) {
                *buf++ = res + 48;
                len++;
                flag = 1;
            }
            num -= res * div;
            div /= 10;
        }
    }
    *buf = '\0';
    return (len);
}

int plcat(char *const dst, const int usiz, const int asiz, const char *fmt,...)
{
    int nlen;
    int flag;
    char *tmp;
    char *s;
    char *pbuf;
    int u;
    va_list args;
    char buf[12];

    if ((asiz == usiz) || (dst == NULL))
        return (0);
    nlen = flag = 0;
    tmp = dst + usiz;
    va_start(args, fmt);
    while(*fmt) {
        if (*fmt == '%')
            flag = 1;
        else if (*fmt == 's') {
            if (flag == 0) {
                nlen++;
                if (asiz > usiz + nlen)
                    *tmp++ = *fmt;
            } else {
                s = va_arg(args, char *);
                nlen += strlen(s);
                if (asiz > usiz + nlen) {
                    while (*s)
                        *tmp++ = *s++;
                }
                flag = 0;
            }
        } else if (*fmt == 'u') {
            if (flag == 0) {
                nlen++;
                if (asiz > usiz + nlen)
                    *tmp++ = *fmt;
            } else {
                u = va_arg(args, int);
                pbuf = buf;
                nlen += itoa(pbuf, u);
                if (asiz > usiz + nlen) {
                    while (*pbuf)
                        *tmp++ = *pbuf++;
                }
                flag = 0;
            }
        } else {
            flag = 0;
            nlen++;
            if (asiz > usiz + nlen)
                *tmp++ = *fmt;
        }
        fmt++;
    }
    va_end(args);
    if (nlen + usiz + 1 > asiz) {
        *(dst + usiz) = 0x00;
        return (-usiz);
    }
    *(dst + usiz + nlen) = 0x00;
    return (usiz + nlen);
}

/*
 * I still keep the old plcat() function here.
 */

#if 0

#include <stdarg.h>
#include <stdio.h>

int plcat(char *const dst, const int usiz, const size_t asiz, const char *fmt,...)
{
    int nlen = 0;
    va_list args;

    va_start(args, fmt);
    nlen = vsnprintf(dst + usiz, asiz - usiz, fmt, args);
    va_end(args);
    if (nlen + usiz + 1 > asiz) {
        *(dst + usiz) = '\0';
        return (-usiz);
    }
    *(dst + usiz + nlen) = '\0';
    return (usiz + nlen);
}

#endif
