static  const char rcsid[] = "$Id: pop3d.c,v 1.25 2005/05/06 12:53:33 dengxf Exp $";

/*
 * Copyright (c) 2003-2004 Deng XueFeng <dsnofe@hotmail.com>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *    This product includes software developed by Deng XueFeng.
 * 4. Neither the name of the author nor the names of any co-contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY DENG XUEFENG AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */

#include <dirent.h>
#include <errno.h>
#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <netinet/in.h>
#include <sys/mman.h>
#include <sys/stat.h>
#include <sys/event.h>
#include <sys/socket.h>
#include <sys/uio.h>
#include <sys/types.h>
#include <arpa/inet.h>

#include "pop3d.h"
#include "const.h"
#include "auth.h"
#include "plcat.h"
#include "mylog.h"

#ifdef EV_SET
/* If event.h is old than 1.22, EV_SET will do the wrong things */
#undef EV_SET
#define EV_SET(kevp_, a, b, c, d, e, f) do {    \
        struct kevent *kevp = (kevp_);      \
        (kevp)->ident = (a);            \
        (kevp)->filter = (b);           \
        (kevp)->flags = (c);            \
        (kevp)->fflags = (d);           \
        (kevp)->data = (e);             \
        (kevp)->udata = (f);            \
} while(0)

#endif
#define Free(x) if (x != NULL) {free(x);x = NULL;}

//#define DEF_MPS_DEBUG   1

static int kq = 0;
static int tsession = 0;
static int online = 0;
static int used_afds = 0;
static int all_cudb = 128;
static int used_audb = 0;
static int used_cudb = 0;
static struct kevent *afds = NULL;
static struct kevent *aevs = NULL;
static UDB *audb = NULL;
static UDB **cudb = NULL;

extern int multi_process;
extern int max_errs;
extern int max_client;
extern int max_timeout;
extern int pop3d_fd;

static int kq_init(void)
{
    kq = kqueue();
    if (kq == -1) {
        INFO("[-ERR]: kq_init: kqueue(2): %s\n", strerror(errno));
        return (-1);
    }
    afds = malloc(max_client * sizeof(struct kevent));
    aevs = malloc(max_client * sizeof(struct kevent));
    audb = malloc(max_client * sizeof(UDB));
    cudb = malloc(all_cudb * sizeof(UDB *));
    if ((afds == NULL) || (aevs == NULL) || (audb == NULL) || (cudb == NULL)) {
        INFO("[-ERR]: kq_init: malloc(3): %s\n", strerror(errno));
        return (-1);
    }
    memset(afds, 0, max_client * sizeof(struct kevent));
    memset(aevs, 0, max_client * sizeof(struct kevent));
    memset(audb, 0, max_client * sizeof(UDB));
    memset(cudb, 0, all_cudb * sizeof(UDB *));
    return (0);
}

static void exitfree(void)
{
    Free(afds);
    Free(aevs);
    Free(audb);
    Free(cudb);
}

static void update_udb(UDB *const addr)
{
    UDB **p;

    if (used_cudb == all_cudb) {
        all_cudb <<= 1;
        p = realloc(cudb, all_cudb * sizeof(UDB *));
        if (p == NULL) {
            INFO("[-ERR]: update_udb: realloc(3): %s\n", strerror(errno));
            exitfree();
            _exit(EXIT_FAILURE);
        }
        cudb = p;
    }
    *(cudb + used_cudb++) = addr;
}

static ssize_t Write(const int fd, const void *const buf, const size_t n)
{
    ssize_t nwritten;

    nwritten = write(fd, buf, n);
    if (nwritten != (int)n) {
        INFO("[-ERR]: Write: write(2): %s\n", strerror(errno));
        return (-1);
    }
    return (nwritten);
}

static ssize_t Awrite(const int fd, const void *const buf, const size_t n)
{
    ssize_t nleft = n;
    ssize_t nwritten;
    const char *ptr = buf;

    while (nleft > 0) {
        if ((nwritten = write(fd, ptr, nleft)) <= 0) {
            INFO("[-ERR]: Awrite: write(2): %s\n", strerror(errno));
            return (-1);
        }
        nleft -= nwritten;
        ptr += nwritten;
    }
    return (n);
}

static void myfree(const struct kevent *const kdp)
{
    UDB *t_udb = kdp->udata;

#ifdef DEF_MPS_DEBUG
    INFO("[DEBUG]: myfree() flag = %d refer = %d\n", t_udb->flag, t_udb->refer);
#endif
    t_udb->cmdlock = DEF_CMD_FREE;
    if (t_udb->flag == DEF_UDB_USED) {
        Free(t_udb->c_all);
        Free(t_udb->c_updated);
    } else if (t_udb->flag == DEF_UDB_INIT) {
        INFO("[EXIT]: myfree() flag = %d refer = %d\n", t_udb->flag, t_udb->refer);
        _exit(EXIT_FAILURE);
    }
    t_udb->flag = DEF_UDB_FREE;
    if (t_udb->refer > 0)
        return;
    close(kdp->ident);
    t_udb->flag = DEF_UDB_INIT;
    update_udb(t_udb);
    online--;
    INFO("[INFO]: Disconnect %s.%u\n",
            inet_ntoa(t_udb->address.sin_addr), htons(t_udb->address.sin_port));
}

static void check_errs(const struct kevent *const kdp)
{
    UDB *t_udb = kdp->udata;

    if (t_udb->curerrs >= max_errs) {
        INFO("[-ERR]: Disconnect %s.%u : Too Many Errors\n",
                inet_ntoa(t_udb->address.sin_addr), htons(t_udb->address.sin_port));
        Awrite(kdp->ident, msg_max_err, sizeof(msg_max_err) - 1);
        myfree(kdp);
    } else {
        EV_SET(&afds[used_afds++], kdp->ident, EVFILT_READ, EV_ADD | EV_ONESHOT, 0, 0, t_udb);
        EV_SET(&afds[used_afds++], kdp->ident, EVFILT_TIMER, EV_ADD | EV_ONESHOT, 0, max_timeout, t_udb);
        t_udb->cmdlock = DEF_CMD_FREE;
        t_udb->refer += 2;
    }
}

static void wel_cmd(const struct kevent *const kdp)
{
    int n;
    UDB *t_udb = kdp->udata;

    n = Awrite(kdp->ident, msg_wel, sizeof(msg_wel) - 1);
    if (n == -1) {
        myfree(kdp);
        return;
    }
    t_udb->welmsg = 1;
    check_errs(kdp);
}

static void unknown_cmd(const struct kevent *const kdp)
{
    int n;
    UDB *t_udb = kdp->udata;

    n = Awrite(kdp->ident, msg_unknown, sizeof(msg_unknown) - 1);
    if (n == -1) {
        myfree(kdp);
        return;
    }
    t_udb->curerrs++;
    t_udb->cmd = 0;
    t_udb->cmderr = 0;
    check_errs(kdp);
}

static void user_cmd(const struct kevent *const kdp)
{
    int n;
    UDB *t_udb = kdp->udata;

    if (t_udb->cmderr != 0) {
        n = Awrite(kdp->ident, msg_user_err, sizeof(msg_user_err) - 1);
        t_udb->curerrs++;
    } else
        n = Awrite(kdp->ident, msg_user_ok, sizeof(msg_user_ok) - 1);
    if (n == -1) {
        myfree(kdp);
        return;
    }
    t_udb->cmd = 0;
    t_udb->cmderr = 0;
    check_errs(kdp);
}

static void pass_cmd(const struct kevent *const kdp)
{
    int n;
    UDB *t_udb = kdp->udata;

    if (t_udb->cmderr == 0)
        n = Awrite(kdp->ident, msg_pass_ok, sizeof(msg_pass_ok) - 1);
    else if (t_udb->cmderr == 1) {
        n = Awrite(kdp->ident, msg_pass_no_user, sizeof(msg_pass_no_user) - 1);
        t_udb->curerrs++;
    } else if (t_udb->cmderr == 2) {
        n = Awrite(kdp->ident, msg_pass_no_pass, sizeof(msg_pass_no_pass) - 1);
        t_udb->curerrs++;
    } else {
        n = Awrite(kdp->ident, msg_pass_err, sizeof(msg_pass_err) - 1);
        t_udb->curerrs++;
    }
    if (n == -1) {
        myfree(kdp);
        return;
    }
    t_udb->cmd = 0;
    t_udb->cmderr = 0;
    check_errs(kdp);
}

static void noop_cmd(const struct kevent *const kdp)
{
    int n;
    UDB *t_udb = kdp->udata;

    n = Awrite(kdp->ident, msg_noop, sizeof(msg_noop) - 1);
    if (n == -1) {
        myfree(kdp);
        return;
    }
    t_udb->cmd = 0;
    t_udb->cmderr = 0;
    check_errs(kdp);
}

static void quit_cmd(const struct kevent *const kdp)
{
    Awrite(kdp->ident, msg_quit, sizeof(msg_quit) - 1);
    myfree(kdp);
}

static void list_cmd(const struct kevent *const kdp)
{
    int i = 0;
    int n;
    int usiz;
    int tmp;
    char outbuf[1024];
    UDB *t_udb = kdp->udata;
    MINFO *t_cache = t_udb->c_all;
    struct iovec iov[2];

    tmp = t_udb->a_id;
    if (tmp == -1) {
        n = Awrite(kdp->ident, msg_id_err, sizeof(msg_id_err) - 1);
        t_udb->cmderr++;
    } else if (tmp > t_udb->maxid) {
        n = Awrite(kdp->ident, msg_id_no, sizeof(msg_id_no) - 1);
        t_udb->cmderr++;
    } else if (t_udb->maxid == 0) {
        n = Awrite(kdp->ident, msg_list_no, sizeof(msg_list_no) - 1);
        t_udb->curerrs++;
    } else if (tmp == 0) {
        strlcpy(outbuf, msg_list_all, sizeof(outbuf));
        usiz = sizeof(msg_list_all) - 1;
        while (i < t_udb->maxid) {
            if ((*(t_cache+i)).f_del != 1)
                usiz = plcat(outbuf, usiz, sizeof(outbuf), "%u %u\r\n", i + 1, (*(t_cache+i)).size);
            if (usiz < 0) {
                n = Awrite(kdp->ident, outbuf, -usiz);
                if (n == -1) {
                    myfree(kdp);
                    return;
                }
                usiz = 0;
                i--;
            }
            i++;
        }
        iov[0].iov_base = outbuf;
        iov[0].iov_len = usiz;
        iov[1].iov_base = (char *)crlfend;
        iov[1].iov_len = sizeof(crlfend) - 1;
        n = writev(kdp->ident, (struct iovec *) &iov, 2);
    } else if ((*(t_cache+tmp-1)).f_del == 1) {
        usiz = plcat(outbuf, 0, sizeof(outbuf), "-ERR message %u is marked as deleted\r\n", tmp);
        n = Awrite(kdp->ident, outbuf, usiz);
        t_udb->cmderr++;
    } else {
        usiz = plcat(outbuf, 0, sizeof(outbuf), "+OK %u %u\r\n", tmp, (*(t_cache+tmp-1)).size);
        n = Awrite(kdp->ident, outbuf, usiz);
    }
    if (n == -1) {
        myfree(kdp);
        return;
    }
    t_udb->cmd = 0;
    t_udb->cmderr = 0;
    check_errs(kdp);
}

static void uidl_cmd(const struct kevent *const kdp)
{
    int i = 0;
    int n;
    int usiz;
    int tmp;
    char outbuf[1024];
    UDB *t_udb = kdp->udata;
    MINFO *t_cache = t_udb->c_all;
    struct iovec iov[2];

    tmp = t_udb->a_id;
    if (tmp == -1) {
        n = Awrite(kdp->ident, msg_id_err, sizeof(msg_id_err) - 1);
        t_udb->cmderr++;
    } else if (tmp > t_udb->maxid) {
        n = Awrite(kdp->ident, msg_id_no, sizeof(msg_id_no) - 1);
        t_udb->cmderr++;
    } else if (t_udb->maxid == 0) {
        n = Awrite(kdp->ident, msg_uidl_no, sizeof(msg_uidl_no) - 1);
        t_udb->curerrs++;
    } else if (tmp == 0) {
        strlcpy(outbuf, msg_ok, sizeof(outbuf));
        usiz = sizeof(msg_ok) - 1;
        while (i < t_udb->maxid) {
            if ((*(t_cache+i)).f_del != 1)
                usiz = plcat(outbuf, usiz, sizeof(outbuf), "%u %s\r\n", i + 1, (*(t_cache+i)).name);
            if (usiz < 0) {
                n = Awrite(kdp->ident, outbuf, -usiz);
                if (n == -1) {
                    myfree(kdp);
                    return;
                }
                usiz = 0;
                i--;
            }
            i++;
        }
        iov[0].iov_base = outbuf;
        iov[0].iov_len = usiz;
        iov[1].iov_base = (char *) crlfend;
        iov[1].iov_len = sizeof(crlfend) - 1;
        n = writev(kdp->ident, (struct iovec *) &iov, 2);
    } else if ((*(t_cache+tmp-1)).f_del == 1) {
        usiz = plcat(outbuf, 0, sizeof(outbuf), "-ERR message %u is marked as deleted\r\n", tmp);
        n = Awrite(kdp->ident, outbuf, usiz);
        t_udb->cmderr++;
    } else {
        usiz = plcat(outbuf, 0, sizeof(outbuf), "+OK %u %s\r\n", tmp, (*(t_cache+tmp-1)).name);
        n = Awrite(kdp->ident, outbuf, usiz);
    }
    if (n == -1) {
        myfree(kdp);
        return;
    }
    t_udb->cmd = 0;
    t_udb->cmderr = 0;
    check_errs(kdp);
}

static int cache_update(UDB *const t_udb, MINFO *const id_updated)
{
    int i = 0;
    MINFO **p_updated;

    if (t_udb->upid == t_udb->ucsize) {
        t_udb->ucsize += 16;
        p_updated = realloc(t_udb->c_updated, t_udb->ucsize * sizeof(MINFO *));
        if (p_updated == NULL) {
            INFO("[-ERR]: cache_update: realloc(3): %s\n", strerror(errno));
            return (-1);
        }
        memset(p_updated+t_udb->upid, 0, 16 * sizeof(MINFO *));
        t_udb->c_updated = p_updated;
    }
    p_updated = t_udb->c_updated;
    while (i < t_udb->upid) {
        if (*(p_updated+i) == id_updated) /* Already Marked */
            return (0);
        i++;
    }
    *(p_updated + t_udb->upid++) = id_updated;
    return (0);
}

static void dele_cmd(const struct kevent *const kdp)
{
    int n;
    int tmp;
    char outbuf[256];
    UDB *t_udb = kdp->udata;
    MINFO *t_cache = t_udb->c_all;

    tmp = t_udb->a_id;
    if (tmp == -1) {
        n = Awrite(kdp->ident, msg_id_err, sizeof(msg_id_err) - 1);
        t_udb->cmderr++;
    } else if ((tmp > t_udb->maxid) || (t_udb->maxid == 0)) {
        n = Awrite(kdp->ident, msg_id_no, sizeof(msg_id_no) - 1);
        t_udb->cmderr++;
    } else if (tmp == 0) {
        n = Awrite(kdp->ident, msg_no_arg, sizeof(msg_no_arg) - 1);
        t_udb->cmderr++;
    } else if ((*(t_cache+tmp-1)).f_del == 1) {
        n = plcat(outbuf, 0, sizeof(outbuf), "-ERR message %u already marked as deleted\r\n", tmp);
        n = Awrite(kdp->ident, outbuf, n);
        t_udb->cmderr++;
    } else {
        (*(t_cache+tmp-1)).f_del = 1;
        n = cache_update(t_udb, t_cache+tmp-1);
        if (n == -1) {
            Awrite(kdp->ident, msg_dele_err, sizeof(msg_dele_err) - 1);
            myfree(kdp);
            return;
        }
        n = plcat(outbuf, 0, sizeof(outbuf), "+OK message %u marked as deleted\r\n", tmp);
        n = Awrite(kdp->ident, outbuf, n);
    }
    if (n == -1) {
        myfree(kdp);
        return;
    }
    t_udb->cmd = 0;
    t_udb->cmderr = 0;
    check_errs(kdp);
}

static void rset_cmd(const struct kevent *const kdp)
{
    int i = 0;
    int n;
    UDB *t_udb = kdp->udata;
    MINFO **p_updated;

    if (t_udb->c_updated != NULL) {
        p_updated = t_udb->c_updated;
        while (i < t_udb->upid) {
            (*(p_updated+i))->f_read = 0;
            (*(p_updated+i))->f_del = 0;
            i++;
        }
        free(t_udb->c_updated);
        t_udb->c_updated = NULL;
    }
    t_udb->ucsize = 0;
    t_udb->upid = 0;
    n = Awrite(kdp->ident, msg_rset, sizeof(msg_rset) - 1);
    if (n == -1) {
        myfree(kdp);
        return;
    }
    t_udb->cmd = 0;
    t_udb->cmderr = 0;
    check_errs(kdp);
}

static void stat_cmd(const struct kevent *const kdp)
{
    int i;
    int curmail = 0;
    int cursize = 0;
    char outbuf[256];
    UDB *t_udb = kdp->udata;
    MINFO *t_cache = t_udb->c_all;

    if (t_udb->maxid != 0) {
        for (i = 0; i < t_udb->maxid; i++) {
            if ((*(t_cache+i)).f_del != 1) {
                cursize += (*(t_cache+i)).size;
                curmail++;
            }
        }
    }
    i = plcat(outbuf, 0, sizeof(outbuf), "+OK %u %u\r\n", curmail, cursize);
    i = Awrite(kdp->ident, outbuf, i);
    if (i == -1) {
        myfree(kdp);
        return;
    }
    t_udb->cmd = 0;
    t_udb->cmderr = 0;
    check_errs(kdp);
}

static void retr_cmd(const struct kevent *const kdp)
{
    int n;
    int tmp;
    int ofd;
    int send_siz;
    int mail_siz;
    char *psend;
    char *mail;
    char sendbuf[def_send_psiz];
    char tmpbuf[256];
    UDB *t_udb = kdp->udata;
    MINFO *t_cache = t_udb->c_all;

    tmp = t_udb->a_id;
    send_siz = (kdp->data > def_send_psiz) ? def_send_psiz : kdp->data;
    if (tmp == -1) {
        n = Awrite(kdp->ident, msg_id_err, sizeof(msg_id_err) - 1);
        t_udb->cmderr++;
        t_udb->cmd = 0;
    } else if (tmp > t_udb->maxid) {
        n = Awrite(kdp->ident, msg_id_no, sizeof(msg_id_no) - 1);
        t_udb->cmderr++;
        t_udb->cmd = 0;
    } else if (tmp == 0) {
        n = Awrite(kdp->ident, msg_no_arg, sizeof(msg_no_arg) - 1);
        t_udb->cmderr++;
        t_udb->cmd = 0;
    } else if ((*(t_cache+tmp-1)).f_del == 1) {
        n = plcat(tmpbuf, 0, sizeof(tmpbuf), "-ERR message %u already marked as deleted\r\n", tmp);
        n = Awrite(kdp->ident, tmpbuf, n);
        t_udb->cmderr++;
        t_udb->cmd = 0;
    } else {
        (*(t_cache+tmp-1)).f_read = 1;
        n = cache_update(t_udb, t_cache+tmp-1);
        if (n == -1) {
            myfree(kdp);
            return;
        }
        if ((t_udb->offset == 0L) && (t_udb->mail == NULL)){
            n = plcat(tmpbuf, 0, sizeof(tmpbuf), "+OK %u octets\r\n", (*(t_cache+tmp-1)).size);
            n = Awrite(kdp->ident, tmpbuf, n);
            if (n == -1) {
                myfree(kdp);
                return;
            }
        }
        if (t_udb->mail == NULL) {
            if ((*(t_cache+tmp-1)).f_new == 1)
                plcat(tmpbuf, 0, sizeof(tmpbuf), "%s/new/%s", t_udb->maildir, (*(t_cache+tmp-1)).name);
            else
                plcat(tmpbuf, 0, sizeof(tmpbuf), "%s/cur/%s", t_udb->maildir, (*(t_cache+tmp-1)).name);
            ofd = open(tmpbuf, O_RDONLY | O_NONBLOCK | O_DIRECT);
            if (ofd < 0) {
                INFO("[-ERR]: retr_cmd: open(2) %s: %s\n", tmpbuf, strerror(errno));
                Awrite(kdp->ident, msg_open_err, sizeof(msg_open_err) - 1);
                myfree(kdp);
                return;
            }
            t_udb->mail = mmap(0, (*(t_cache+tmp-1)).size, PROT_READ, MAP_PRIVATE, ofd, 0L);
            close(ofd);
            if (t_udb->mail == MAP_FAILED) {
                INFO("[-ERR]: retr_cmd: mmap(2): %s\n", strerror(errno));
                Awrite(kdp->ident, msg_mmap_err, sizeof(msg_mmap_err) - 1);
                myfree(kdp);
                return;
            }
        }
        psend = sendbuf;
        mail = t_udb->mail + t_udb->offset;
        mail_siz = (*(t_cache+tmp-1)).size;
        while ((psend - sendbuf + 1) <= send_siz - 1) {
            if (mail >= t_udb->mail + mail_siz)
                break;
            if (*mail == '\n') {
                *psend++ = '\r';
                *psend++ = '\n';
            } else if ((*mail  == '.') && (*(mail-1) == '\n')) {
                *psend++ = '.';
                *psend++ = '.';
            } else
                *psend++ = *mail;
            mail++;
        }
        if (((psend - sendbuf) < send_siz) && (mail < t_udb->mail + mail_siz)) {
            if (*mail == '\n')
                *psend = '\0';
            else if ((*mail  == '.') && (*(mail-1) == '\n'))
                *psend = '\0';
            else
                *psend++ = *mail;
            mail++;
        }
        n = Write(kdp->ident, sendbuf, psend - sendbuf);
        if (n == -1) {
            myfree(kdp);
            return;
        }
        t_udb->offset = mail - t_udb->mail;
        if (t_udb->offset == mail_siz) {
            n = Awrite(kdp->ident, crlfend, sizeof(crlfend) - 1);
            n = munmap(t_udb->mail, mail_siz);
            if (n == -1) {
                INFO("[-ERR]: retr_cmd: munmap(2): %s\n", strerror(errno));
                Awrite(kdp->ident, msg_munmap_err, sizeof(msg_munmap_err) - 1);
                myfree(kdp);
                return;
            }
            t_udb->cmd = 0;
            t_udb->mail = NULL;
            t_udb->offset = 0L;
        }
    }
    if (n == -1) {
        myfree(kdp);
        return;
    }
    if (t_udb->cmd == 0) {
        t_udb->cmderr = 0;
        check_errs(kdp);
    } else {
        EV_SET(&afds[used_afds++], kdp->ident, EVFILT_WRITE, EV_ADD | EV_ONESHOT, 0, 0, t_udb);
        t_udb->refer++;
    }
}

static void exit_cmd(const struct kevent *const kdp)
{
    int i;
    int res;
    char fdir[256];
    char tdir[256];
    UDB *t_udb = kdp->udata;
    MINFO **p_updated;

    if (t_udb->ucsize != 0) {
        p_updated = t_udb->c_updated;
        for (i=0; i<t_udb->upid; i++) {
            if ((*(p_updated+i))->f_del == 1) {
                if ((*(p_updated+i))->f_new == 1)
                    plcat(fdir, 0, sizeof(fdir), "%s/new/%s", t_udb->maildir, (*(p_updated+i))->name);
                else
                    plcat(fdir, 0, sizeof(fdir), "%s/cur/%s", t_udb->maildir, (*(p_updated+i))->name);
                res = unlink(fdir);
                if (res != 0) {
                    INFO("[-ERR]: exit_cmd: unlink(2) %s: %s\n", fdir, strerror(errno));
                    Awrite(kdp->ident, msg_unlink_err, sizeof(msg_unlink_err) - 1);
                    myfree(kdp);
                    return;
                }
            } else if ((*(p_updated+i))->f_read == 1) {
                if ((*(p_updated+i))->f_new == 1) {
                    plcat(fdir, 0, sizeof(fdir), "%s/new/%s", t_udb->maildir, (*(p_updated+i))->name);
                    plcat(tdir, 0, sizeof(tdir), "%s/cur/%s", t_udb->maildir, (*(p_updated+i))->name);
                    res = rename(fdir, tdir);
                    if (res != 0) {
                        INFO("[-ERR]: exit_cmd: rename(2) %s => %s: %s\n", fdir, tdir, strerror(errno));
                        Awrite(kdp->ident, msg_rename_err, sizeof(msg_rename_err) - 1);
                        myfree(kdp);
                        return;
                    }
                }
            }
        }
    }
    i = Awrite(kdp->ident, msg_quit, sizeof(msg_quit) - 1);
    if (i == -1) {
        myfree(kdp);
        return;
    }
    myfree(kdp);
}

static void tmperr_cmd(const struct kevent *const kdp)
{
    int n;
    UDB *t_udb = kdp->udata;

    n = Awrite(kdp->ident, msg_opendir_err, sizeof(msg_opendir_err) - 1);
    if (n == -1) {
        myfree(kdp);
        return;
    }
    t_udb->cmd = 0;
    t_udb->cmderr = 0;
    check_errs(kdp);
}

static void cmd_write(const struct kevent *const kdp)
{
    UDB *w_udb = kdp->udata;

    w_udb->refer--;
    if (w_udb->flag == DEF_UDB_FREE) {
        myfree(kdp);
        return;
    }
    if (w_udb->welmsg == 0)
        wel_cmd(kdp);
    else if (w_udb->cmd == DEF_CMDNAME_UNKNOWN)
        unknown_cmd(kdp);
    else if (w_udb->cmd == DEF_CMDNAME_USER)
        user_cmd(kdp);
    else if (w_udb->cmd == DEF_CMDNAME_PASS)
        pass_cmd(kdp);
    else if (w_udb->cmd == DEF_CMDNAME_QUIT)
        quit_cmd(kdp);
    else if (w_udb->cmd == DEF_CMDNAME_RSET)
        rset_cmd(kdp);
    else if (w_udb->cmd == DEF_CMDNAME_NOOP)
        noop_cmd(kdp);
    else if (w_udb->cmd == DEF_CMDNAME_EXIT)
        exit_cmd(kdp);
    else if (w_udb->cache_status == DEF_CACHE_UNDONE)
        tmperr_cmd(kdp);
    else if (w_udb->cmd == DEF_CMDNAME_LIST)
        list_cmd(kdp);
    else if (w_udb->cmd == DEF_CMDNAME_UIDL)
        uidl_cmd(kdp);
    else if (w_udb->cmd == DEF_CMDNAME_DELE)
        dele_cmd(kdp);
    else if (w_udb->cmd == DEF_CMDNAME_RETR)
        retr_cmd(kdp);
    else if (w_udb->cmd == DEF_CMDNAME_STAT)
        stat_cmd(kdp);
    else
        unknown_cmd(kdp);
}

static char *upch(char *const str)
{
    char *tmp;

    tmp = str;
    while(*tmp != '\0') {
        if ((*tmp > 'z') || (*tmp < 'A'))
            return (NULL);
        else if ((*tmp > 'Z') && (*tmp < 'a'))
            return (NULL);
        else if ((*tmp >= 'a') && (*tmp <= 'z'))
            *tmp -= 32;
        tmp++;
    }
    return (str);
}

static int aldigit(char *const data, const int maxn)
{
    int i;
    char *t;

    if (*data == '\0')
        return (0);
    t = data;
    while (*t != '\0') {
        if ((*t < '0') || (*t > '9'))
            return (-1);
        t++;
    }
    i = atoi(data);
    return (i == 0) ? (maxn + 3) : (i);
}

static int cmd_parse(CMDTBL *const p_cmd, char *cmdbuf)
{
    int flags;
    int values;
    char *tmp;

    tmp = cmdbuf + 4;
    if (*tmp != '\r' && *tmp != ' ')
        return (0);
    *tmp++ = '\0';
    cmdbuf = upch(cmdbuf);
    if (cmdbuf == NULL)
        return (0);
    strcpy(p_cmd->name, cmdbuf);
    if (*tmp == '\n')
        return (1);
    flags = 0;
    values = 0;
    while(*tmp != '\0') {
        if ((*tmp == '\r') && (*(tmp + 1) == '\n')) {
            *tmp = '\0';
            flags = 1;
            break;
        } else if ((values == 0) && (*tmp != ' '))
            values = 1;
        tmp++;
    }
    if (flags == 0)
        return (-1);
    if (values == 1) {
        strcpy(p_cmd->value, cmdbuf + 5);
        return (2);
    } else
        return (1);
}

static int md_cache(const char *dir, int siz, const int new, UDB *const m_udb)
{
    int id;
    char tdir[256];
    struct dirent *dp;
    struct stat stat_buf;
    DIR *dirp;
    MINFO *p;

    if (m_udb->c_all == NULL) {
        p = malloc(siz * sizeof(MINFO));
        if (p == NULL) {
            INFO("[-ERR]: md_cache: malloc(3): %s\n", strerror(errno));
            return (-1);
        }
        m_udb->c_all = p;
        memset(p, 0, siz * sizeof(MINFO));
    }
    if ((dirp = opendir(dir)) == NULL) {
        INFO("[-ERR]: md_cache: opendir(2) %s: %s\n", dir, strerror(errno));
        return (-1);
    }
    p = m_udb->c_all;
    id = m_udb->maxid;
    while ((dp = readdir(dirp)) != NULL) {
        if (id == (siz - 1)) {
            siz <<= 1;
            p = realloc(m_udb->c_all, siz * sizeof(MINFO));
            if (p == NULL) {
                INFO("[-ERR]: md_cache: realloc(3): %s\n", strerror(errno));
                closedir(dirp);
                return (-1);
            }
            m_udb->c_all = p;
        }
        if (((strcmp(".", dp->d_name)) == 0) || ((strcmp("..", dp->d_name)) == 0))
            continue;
        plcat(tdir, 0, sizeof(tdir), "%s/%s", dir, dp->d_name);
        if ((stat(tdir, &stat_buf)) == -1) {
            INFO("[-ERR]: md_cache: stat(2) %s: %s\n", tdir, strerror(errno));
            closedir(dirp);
            return (-1);
        }
        (*(p+id)).id = id + 1;
        (*(p+id)).f_del = 0;
        (*(p+id)).f_read = 0;
        (*(p+id)).f_new = new;
        strlcpy((*(p+id)).name, dp->d_name, sizeof(p->name));
        (*(p+id)).size = stat_buf.st_size;
        id++;
    }
    m_udb->maxid = id;
    if ((closedir(dirp)) != 0) {
        INFO("[-ERR]: md_cache: closedir(3) %s: %s\n", dirp, strerror(errno));
        return (-1);
    }
    return (siz);
}

static void init_cache(UDB *const g_udb)
{
    int res;
    char tmpdir[256];

    plcat(tmpdir, 0, sizeof(tmpdir), "%s/new", g_udb->maildir);
    res = md_cache(tmpdir, 32, 1, g_udb);
    if (res == -1)
        return;
    plcat(tmpdir, 0, sizeof(tmpdir), "%s/cur", g_udb->maildir);
    res = md_cache(tmpdir, res, 0, g_udb);
    if (res == -1)
        return;
    g_udb->cache_status = DEF_CACHE_DONE;
}

static void cmd_read(const struct kevent *const kdp)
{
    int n;
    char inbuf[256];
    UDB *r_udb = kdp->udata;
    CMDTBL r_cmd;

    EV_SET(&afds[used_afds++], kdp->ident, EVFILT_TIMER, EV_DELETE, 0, 0, NULL);
    r_udb->cmdlock = DEF_CMD_LOCK;
    r_udb->refer -= 2;
    if (r_udb->flag == DEF_UDB_FREE) {
        myfree(kdp);
        return;
    }
    n = read(kdp->ident, &inbuf, sizeof(inbuf)-1);
    if (n <= 0) {
        INFO("[-ERR]: cmd_read: read(2): %s\n", strerror(errno));
        myfree(kdp);
        return;
    }
    *(inbuf+n) = '\0';
    memset(&r_cmd, 0, sizeof(CMDTBL));
    n = cmd_parse(&r_cmd, inbuf);
    if (n == -1) {
        Awrite(kdp->ident, msg_over, sizeof(msg_over) - 1);
        myfree(kdp);
        return;
    } else if (n == 0) {
        r_udb->cmd = DEF_CMDNAME_UNKNOWN;
        r_udb->cmderr = 0;
    } else if (r_udb->login == 0) {
        if (strcmp(r_cmd.name, "USER") == 0) {
            r_udb->cmd = DEF_CMDNAME_USER;
            if (n == 2) {
                plcat(r_udb->username, 0, sizeof(r_udb->username), "%s", r_cmd.value);
                r_udb->cmderr = 0;
            } else
                r_udb->cmderr = 1;
        } else if (strcmp(r_cmd.name, "PASS") == 0) {
            r_udb->cmd = DEF_CMDNAME_PASS;
            if (r_udb->username[0] == '\0')
                r_udb->cmderr = 1;
            else if (n == 1)
                r_udb->cmderr = 2;
            else {
                n = auth_system(r_udb->username, r_udb->maildir, sizeof(r_udb->maildir), r_cmd.value);
                memset(&r_cmd, 0, sizeof(CMDTBL));
                if (n == 1) {
                    r_udb->cmderr = 0;
                    r_udb->login = 1;
                } else {
                    INFO("[-ERR]: Auth Error FROM: %s.%u : user: %s\n",
                            inet_ntoa(r_udb->address.sin_addr), htons(r_udb->address.sin_port), r_udb->username);
                    r_udb->cmderr = 3;
                }
            }
        } else if (strcmp(r_cmd.name, "QUIT") == 0) {
            r_udb->cmd = DEF_CMDNAME_QUIT;
            r_udb->cmderr = 0;
        } else
            r_udb->cmd = 0;
    } else {
        if (strcmp(r_cmd.name, "LIST") == 0) {
            r_udb->cmd = DEF_CMDNAME_LIST;
            r_udb->a_id = aldigit(r_cmd.value, r_udb->maxid);
        } else if (strcmp(r_cmd.name, "UIDL") == 0) {
            r_udb->cmd = DEF_CMDNAME_UIDL;
            r_udb->a_id = aldigit(r_cmd.value, r_udb->maxid);
        } else if (strcmp(r_cmd.name, "DELE") == 0) {
            r_udb->cmd = DEF_CMDNAME_DELE;
            r_udb->a_id = aldigit(r_cmd.value, r_udb->maxid);
        } else if (strcmp(r_cmd.name, "RETR") == 0) {
            r_udb->cmd = DEF_CMDNAME_RETR;
            r_udb->a_id = aldigit(r_cmd.value, r_udb->maxid);
        } else if (strcmp(r_cmd.name, "STAT") == 0) {
            r_udb->cmd = DEF_CMDNAME_STAT;
            r_udb->cmderr = 0;
        } else if (strcmp(r_cmd.name, "RSET") == 0) {
            r_udb->cmd = DEF_CMDNAME_RSET;
            r_udb->cmderr = 0;
        } else if (strcmp(r_cmd.name, "NOOP") == 0) {
            r_udb->cmd = DEF_CMDNAME_NOOP;
            r_udb->cmderr = 0;
        } else if (strcmp(r_cmd.name, "QUIT") == 0) {
            r_udb->cmd = DEF_CMDNAME_EXIT;
            r_udb->cmderr = 0;
        } else {
            r_udb->cmd = DEF_CMDNAME_UNKNOWN;
            r_udb->cmderr = 0;
        }
        if ((r_udb->cmd < DEF_CMDNAME_RSET) && (r_udb->cache_status == DEF_CACHE_UNDONE))
            init_cache(r_udb);
    }
    EV_SET(&afds[used_afds++], kdp->ident, EVFILT_WRITE, EV_ADD | EV_ONESHOT, 0, 0, r_udb);
    r_udb->refer++;
}

static void checktm(const struct kevent *const kdp)
{
    UDB *t_udb = kdp->udata;

    if (t_udb->cmdlock == DEF_CMD_LOCK)
        return;
    t_udb->refer--;
    if (t_udb->flag == DEF_UDB_USED)
        Awrite(kdp->ident, msg_timeout, sizeof(msg_timeout) - 1);
    shutdown(kdp->ident, SHUT_RD);
    myfree(kdp);
}

static void cli_accept(const struct kevent *const kdp)
{
    int cli_fd;
    UDB *udbp;
    socklen_t tmpaddrlen;
    struct sockaddr_in address;

    tmpaddrlen = sizeof(struct sockaddr_in);
    cli_fd = accept(kdp->ident, (struct sockaddr *) &address, &tmpaddrlen);
    if (cli_fd < 0) {
        INFO("[-ERR]: cli_accept: accept(2): %s\n", strerror(errno));
        return;
    } else {
        tsession++;
        if (online == max_client) {
            INFO("[-ERR]: Total:%d Online: %d ACCEPT FROM: %s.%u : Too Many Online\n",
                    tsession, online, inet_ntoa(address.sin_addr), htons(address.sin_port));
            Write(cli_fd, msg_max_online, sizeof(msg_max_online) - 1);
            close(cli_fd);
            return;
        }
        online++;
        INFO("%d [INFO]: Total:%d Online: %d ACCEPT FROM: %s.%u\n",
                getpid(), tsession, online, inet_ntoa(address.sin_addr), htons(address.sin_port));
        udbp = (used_cudb == 0) ? &audb[used_audb++] : cudb[--used_cudb];
        memset(udbp, 0, sizeof(UDB));
        udbp->flag = DEF_UDB_USED;
        memcpy(&udbp->address, &address, sizeof(address));
        EV_SET(&afds[used_afds++], cli_fd, EVFILT_WRITE, EV_ADD | EV_ONESHOT, 0, 0, udbp);
        udbp->refer = 1;
    }
}

static void myloop(int id)
{
    int myid;
    int res;
    struct kevent *kdp;
    UDB *t_udb = NULL;

    myid = 0;
    while (1) {
        errno = 0;
        res = kevent(kq, afds, used_afds, aevs, max_client, NULL);
        if (res == -1) {
            INFO("[-ERR]: myloop: kevent(2): %s\n", strerror(errno));
            continue;
        }
        if (res == 0) {
            INFO("[EXIT]: myloop: NO Event received\n");
            exitfree();
            _exit(EXIT_FAILURE);
        }
        used_afds = 0;
        for (kdp = aevs; kdp < &aevs[res]; kdp++) {
            if ((int)kdp->ident != pop3d_fd)
                t_udb = kdp->udata;
            if ((int)kdp->ident == pop3d_fd) {
                if (kdp->filter == EVFILT_READ)
                    cli_accept(kdp);
                else {
                    if (myid == id)
                        EV_SET(&afds[used_afds++], pop3d_fd, EVFILT_READ, EV_ENABLE, 0, 0, NULL);
                    else
                        EV_SET(&afds[used_afds++], pop3d_fd, EVFILT_READ, EV_DISABLE, 0, 0, NULL);
                    if (myid == multi_process-1)
                        myid = 0;
                    else
                        myid++;
                }
            } else if (kdp->flags & EV_ERROR) {
                if ((int)kdp->ident != pop3d_fd) {
                    t_udb->refer--;
                    myfree(kdp);
                }
            } else if (kdp->flags & EV_EOF) {
                if ((int)kdp->ident != pop3d_fd) {
                    t_udb->refer--;
                    myfree(kdp);
                }
            } else if (kdp->filter == EVFILT_TIMER)
                checktm(kdp);
            else if (kdp->filter == EVFILT_READ)
                cmd_read(kdp);
            else if (kdp->filter == EVFILT_WRITE)
                cmd_write(kdp);
            else {
                INFO("[-ERR]: myloop: UNKNOWN\n");
                exitfree();
                _exit(EXIT_FAILURE);
            }
        }
    }
}

void pop3d(int id)
{
    if (kq_init() == -1) {
        exitfree();
        _exit(EXIT_FAILURE);
    }
    if (multi_process == 1)
        EV_SET(&afds[used_afds++], pop3d_fd, EVFILT_READ, EV_ADD, 0, 0, NULL);
    else {
        EV_SET(&afds[used_afds++], pop3d_fd, EVFILT_READ, EV_ADD|EV_DISABLE, 0, 0, NULL);
        EV_SET(&afds[used_afds++], pop3d_fd, EVFILT_TIMER, EV_ADD, 0, 500, NULL);
    }
    myloop(id);
}
